@@ -3,7 +3,7 @@
 # -*- coding: utf-8 -*-
 # create-upstream-tarballs - Utility to create the required source tarballs
 #                            to package Thunderbird within Debian
-# Copyright (c) 2022-2024 Carsten Schoenert <c.schoenert@t-online.de>
+# Copyright (c) 2022-2025 Carsten Schoenert <c.schoenert@t-online.de>
 #
 # SPDX-License-Identifier: GPL-2.0-or-later
 
@@ -33,26 +33,8 @@ except ImportError:
 # local stuff
 import repack
 
-TB_BASE_URL_RELEASES = "https://ftp.mozilla.org/pub/thunderbird/releases/"
-TB_BASE_URL_CANDIDATES = "https://ftp.mozilla.org/pub/thunderbird/candidates/"
-
-# Some strings we need to exclude within the detection of the current
-# release version.
-RELEASE_STRING_EXCLUSION_LIST = (
- "-real",
- "125.0",
- "127.0",
- "129.0",
-)
-
-# Also we need to exclude some strings in the current planned version
-# detection.
-PLANNED_STRING_EXCLUSION_LIST = (
- "124.0",
- "125.0",
- "127.0",
- "129.0",
-)
+TB_BASE_URL_RELEASES = "https://download-origin.cdn.mozilla.net/pub/thunderbird/releases/"
+TB_BASE_URL_CANDIDATES = "https://download-origin.cdn.mozilla.net/pub/thunderbird/candidates/"
 
 try:
     import colorlog
@@ -101,7 +83,10 @@ def option_parser():
     )
 
     parser.add_argument(
-        "-d", "--debug", action="store_true", help="Enable output of debug messages."
+        "-d",
+        "--debug",
+        action="store_true",
+        help="Enable output of debug messages."
     )
 
     parser.add_argument(
@@ -202,40 +187,77 @@ def get_versions():
 
     if tb_release_candidates_html != "None":
         tb_href_entries = tb_release_candidates_html.xpath("//a /@href")
-        # Variable to include all found planned ESR versions.
-        tb_upstream_candidates_versions = []
+        tb_upstream_candidates_non_versions = [] # List for all non ESR candidates versions.
+        tb_upstream_candidates_esr_versions = [] # List for all ESR candidates versions.
+
         for elem in tb_href_entries:
             # Processing all href elements and collect all of them with a numeric version.
             line = elem.split("/")
+            # Loop through all ESR versions.
             if (
                 len(line) == 6
                 and line[3] == "candidates"
-                and line[4].endswith("candidates")
-                # Exclude all "wrong" versions which never will become a ESR version.
-                and not line[4].startswith(PLANNED_STRING_EXCLUSION_LIST)
-                and "b" not in line[4]
+                and "esr" in line[4]
             ):
                 # We have now all versions with '-candidates' as suffix which
-                # need to get split of.
-                tb_upstream_candidates_versions.append(line[4].split("-")[0])
+                # need to get split off and drop the 'esr' suffix.
+                tb_upstream_candidates_esr_versions.append(line[4].split("-")[0].removesuffix("esr"))
+
+            # Loop through all non ESR versions.
+            elif (
+                len(line) == 6
+                and line[3] == "candidates"
+                and "b" not in line[4]
+            ):
+                tb_upstream_candidates_non_versions.append(line[4].split("-")[0])
+
+        tb_upstream_candidates_esr_versions.sort(key=lambda version: tuple(map(int, version.split('.'))))
+        tb_upstream_candidates_non_versions.sort(key=lambda version: tuple(map(int, version.split('.'))))
+
+        # Step 1.1 - Sorting the ESR candidates versions.
+        #
+        # Creating a list which is containing then again multiple lists of
+        # major ESR versions.
+        recent_major = 0
+        tmp_major = 0
+        tmp_esr_list = []
+
+        for version in tb_upstream_candidates_esr_versions:
+            recent_major = int(version.split(".")[0])
+
+            if tmp_major < recent_major:
+                tmp_major = recent_major
+                tmp_esr_list.append([version])
+            else:
+                index = len(tmp_esr_list) - 1
+                tmp_esr_list[index].extend([version])
+
 
     else:
         logging.error("Could not load versions from '%s'!", TB_BASE_URL_CANDIDATES)
         sys.exit(1)
 
-    release_planned_version = tb_upstream_candidates_versions[1]
-    for pos in range(2, len(tb_upstream_candidates_versions)):
-        check = Version(release_planned_version.replace('esr', '')) < Version(
-            tb_upstream_candidates_versions[pos].replace('esr', '')
-        )
-        if check:
-            release_planned_version = tb_upstream_candidates_versions[pos]
-
-    # Now searching the newest build for the found candidates version.
-    release_candidate_build = get_latest_candidates_build(
-        f"{TB_BASE_URL_CANDIDATES}{release_planned_version}-candidates/"
+    # Step 1.2 - Picking the recent ESR candidates versions.
+    #
+    # Getting the recent candidate version for the old (previous) ESR version.
+    # Take the last element from the second last list.
+    release_planned_old_esr_version = f"{tmp_esr_list[-2][-1]}esr"
+    # And the build for this version.
+    release_planned_old_esr_version_build = get_latest_candidates_build(
+        f"{TB_BASE_URL_CANDIDATES}{release_planned_old_esr_version}-candidates/"
     )
-    release_candidate_version = release_planned_version
+
+    # Getting the recent candidate version for the new ESR version.
+    # Take the last element from the last list.
+    release_planned_next_esr_version = f"{tmp_esr_list[-1][-1]}esr"
+    # And the build for this version.
+    release_planned_next_esr_version_build = get_latest_candidates_build(
+        f"{TB_BASE_URL_CANDIDATES}{release_planned_next_esr_version}-candidates/"
+    )
+
+    # Step 1.3 - Picking the most recent non ESR candidate version.
+    #
+    candidate_non_esr_version = tb_upstream_candidates_non_versions[-1]
 
     # Step 2 - Get data about released ESR and Beta versions
     #
@@ -246,7 +268,7 @@ def get_versions():
     if tb_releases_html != "None":
         tb_href_entries = tb_releases_html.xpath("//a /@href")
         # Variable to include all released ESR versions.
-        tb_upstream_versions = []
+        tb_upstream_release_versions = []
         for elem in tb_href_entries:
             # Processing all href elements and collect all of them with a numeric version.
             line = elem.split("/")
@@ -257,8 +279,8 @@ def get_versions():
                     "custom-updates",
                     "latest",
                     "latest-beta",
-                ] and not line[4].endswith(RELEASE_STRING_EXCLUSION_LIST):
-                    tb_upstream_versions.append(line[4])
+                ]:
+                    tb_upstream_release_versions.append(line[4])
 
     else:
         logging.error("Could not load versions from '%s'!", TB_BASE_URL_RELEASES)
@@ -266,31 +288,110 @@ def get_versions():
 
     # Step 3 - Find most recent Release and Beta versions.
     #
-    release_version = tb_upstream_versions[1]
-    beta_version = release_version
-    for pos in range(2, len(tb_upstream_versions)):
-        if "b" not in tb_upstream_versions[pos]:
-            check = Version(release_version.replace('esr', '')) < Version(
-                tb_upstream_versions[pos].replace('esr', '')
-            )
-            if check:
-                release_version = tb_upstream_versions[pos]
+    # Step 3.1 - Create (three) lists for each version type.
+    #
+    # One for ESR Releases "*esr", one for Release "beta" versions and one
+    # for the non ESR versions.
+    esr_release_versions = []
+    beta_release_versions = []
+    release_non_esr_versions = []
+    for version in tb_upstream_release_versions:
+        if "esr" in str(version):
+            esr_release_versions.append(version.removesuffix("esr"))
 
-        if "b" in tb_upstream_versions[pos]:
+        elif "b" in str(version):
+            beta_release_versions.append(version)
+
+        elif (
+            "rc" not in str(version) and
+            "a" not in str(version) and
+            "real" not in str(version)
+        ):
+            release_non_esr_versions.append(version)
+
+    # Step 3.2 - Sort the Release "*esr" list and the non ESR releases by
+    # numeric numbers ascending.
+    #
+    esr_release_versions.sort(key=lambda version: tuple(map(int, version.split('.'))))
+    release_non_esr_versions.sort(key=lambda version: tuple(map(int, version.split('.'))))
+
+    # Step 3.3 - Get the most previous and current Release ESR versions.
+    #
+    recent_major = 0
+    tmp_major = 0
+    tmp_release_list = []
+
+    for version in esr_release_versions:
+        recent_major = int(version.split(".")[0])
+
+        if tmp_major < recent_major:
+            tmp_major = recent_major
+            tmp_release_list.append([version])
+        else:
+            index = len(tmp_release_list) - 1
+            tmp_release_list[index].extend([version])
+
+    # Picking the recent version for the old (previous) ESR version.
+    # Take the last element from the second last list.
+    release_prev_esr_version = f"{tmp_release_list[-2][-1]}esr"
+
+    # Picking the recent version for the new ESR version.
+    # Take the last element from the last list.
+    release_cur_esr_version = f"{tmp_release_list[-1][-1]}esr"
+
+    # Step 3.4 - Get the newest "beta" version.
+    #
+    beta_version = tb_upstream_release_versions[1]
+    for pos in range(2, len(beta_release_versions)):
+        if "b" in beta_release_versions[pos]:
             check = Version(beta_version) < Version(
-                tb_upstream_versions[pos]
+                beta_release_versions[pos]
             )
             if check:
-                beta_version = tb_upstream_versions[pos]
+                beta_version = beta_release_versions[pos]
 
-    logging.debug("Current Release version: %s", release_version)
-    logging.debug("Current (planned) Release version: %s (build%s)", release_candidate_version, release_candidate_build)
+    # Step 4 - Find the most recent non ESR versions.
+    #
+    release_non_esr_version = release_non_esr_versions[-1]
+
+    logging.debug(
+        "Current Release ESR (previous) version: %s",
+        release_prev_esr_version
+    )
+    logging.debug(
+        "Current Release ESR (current) version: %s",
+        release_cur_esr_version
+    )
+    logging.debug(
+        "Current Release non ESR version: %s",
+        release_non_esr_version
+    )
+    logging.debug(
+        "Current Candidate ESR (previous) version: %s (build%s)",
+        release_planned_old_esr_version,
+        release_planned_old_esr_version_build
+    )
+    logging.debug(
+        "Current Candidate ESR (current) version: %s (build%s)",
+        release_planned_next_esr_version,
+        release_planned_next_esr_version_build
+    )
+    logging.debug(
+        "Current Candidate non ESR version: %s",
+        candidate_non_esr_version
+    )
     logging.debug("Current Beta version: %s", beta_version)
+
     return [
-        release_version,
-        release_candidate_version,
-        release_candidate_build,
+        release_prev_esr_version,
+        release_cur_esr_version,
+        release_planned_old_esr_version,
+        release_planned_old_esr_version_build,
+        release_planned_next_esr_version,
+        release_planned_next_esr_version_build,
         beta_version,
+        release_non_esr_version,
+        candidate_non_esr_version,
     ]
 
 
@@ -300,6 +401,7 @@ def get_xpi_languages(base_url):
     """
     logging.debug("Using %s to search for xpi files.", base_url)
     xpi_html = get_website_data(base_url)
+
     # Try to get xpath data elements, will be successful if the URL is a
     # valid and usable URL for the requested version! Version to download
     # which are set manually will always be looked for on the release channel.
@@ -316,12 +418,12 @@ def get_xpi_languages(base_url):
         # ['', 'pub', 'thunderbird', 'candidates', '$(version)-candidates', 'build1', 'linux-x86_64', 'xpi', '$(lang).xpi']
         # We can ignore the en-US language, as this is the native language of TB.
 
-        # Catch up XPI data from release channel
+        # Catch up XPI data from release channel.
         if len(line) == 8 and line[3] == "releases":
             if not line[7].endswith("en-US.xpi"):
                 xpi_list.append(line[7])
 
-        # Catch up XPI data from candidates channel
+        # Catch up XPI data from candidates channel.
         if len(line) == 9 and line[3] == "candidates":
             if not line[8].endswith("en-US.xpi"):
                 xpi_list.append(line[8])
@@ -336,33 +438,53 @@ def get_xpi_languages(base_url):
 
 def select_upstream_version(detected_upstream_versions):
     """
-    Gives back the upstream version the user wants to use.
+    Gives back the upstream version the user wants to proceed.
     """
     while True:
         print("\nPlease select the Thunderbird version to download!")
-        print(f"1 - Current Release version: {detected_upstream_versions[0]}")
-        print(
-            f"2 - Current Planned version: {detected_upstream_versions[1]} build{detected_upstream_versions[2]}"
-        )
-        print(f"3 - Current Beta version   : {detected_upstream_versions[3]}")
-        print("4 - Do nothing and Exit\n")
+        print("------------------------------------------")
+        print(f"1 - Current ESR Release (previous) version: {detected_upstream_versions[0]}")
+        print(f"2 - Current ESR Release (current) version:  {detected_upstream_versions[1]}")
+        print("------------------------------------------")
+        print(f"3 - Current Planned ESR (previous) version: {detected_upstream_versions[2]} build{detected_upstream_versions[3]}")
+        print(f"4 - Current Planned ESR (current) version:  {detected_upstream_versions[4]} build{detected_upstream_versions[5]}")
+        print("------------------------------------------")
+        print(f"5 - Current Beta version:                   {detected_upstream_versions[6]}")
+        print(f"6 - Current non ESR Release version:        {detected_upstream_versions[7]}")
+        print(f"7 - Current non ESR Candidate version:      {detected_upstream_versions[8]}")
+        print("------------------------------------------")
+        print("8 - Do nothing and Exit\n")
         print("Your selection: ", end=" ")
-        version = int(input())
-        if version not in [1, 2, 3, 4]:
-            print("Wrong input! Please select a valid number for choosing.")
-        elif version == 4:
-            logging.info("Action aborted by the user.")
+        try:
+            version = int(input())
+            if version not in range(1, 9):
+                print("\u001b[31mWrong input!\u001b[0m Please select a valid number for choosing.")
+            elif version == 8:
+                logging.info("Action aborted by the user.")
+                sys.exit(1)
+            else:
+                print()
+                break
+        except ValueError:
+            print("\u001b[31mWrong input!\u001b[0m Please select a valid number for choosing.")
+        except KeyboardInterrupt:
+            print("   \u001b[31mAborted!\u001b[0m")
             sys.exit(1)
-        else:
-            print()
-            break
 
     if version == 1:
         return detected_upstream_versions[0]
     elif version == 2:
-        return [detected_upstream_versions[1], detected_upstream_versions[2]]
-    else:
-        return detected_upstream_versions[3]
+        return detected_upstream_versions[1]
+    elif version == 3:
+        return [detected_upstream_versions[2], detected_upstream_versions[3]]
+    elif version == 4:
+        return [detected_upstream_versions[4], detected_upstream_versions[5]]
+    elif version == 5:
+        return detected_upstream_versions[6]
+    elif version == 6:
+        return detected_upstream_versions[7]
+    elif version == 7:
+        return detected_upstream_versions[8]
 
 
 def download_file(url, file, target_folder):
@@ -440,6 +562,7 @@ def create_tb_l10n_tarball(xpi_languages, base_folder, upstream_version):
     """
     logging.info("Prepare data for l10n component tarball.")
     l10n_upstream_folder = f"{base_folder}/thunderbird-l10n"
+
     if os.path.isdir(l10n_upstream_folder):
         if len([entry for entry in os.listdir(l10n_upstream_folder)]) > 0:
             for entry in os.scandir(l10n_upstream_folder):
@@ -565,13 +688,34 @@ if __name__ == "__main__":
     if args.get_versions:
         upstream_versions = get_versions()
         logging.info(
-            "Current most recent TB Release ESR version:       %s", upstream_versions[0]
+            "Current most recent TB Release ESR (previous) version:   %s",
+            upstream_versions[0]
         )
         logging.info(
-            "Current most recent TB Release candidate version: %s build%s", upstream_versions[1], upstream_versions[2]
+            "Current most recent TB Release ESR (current) version:    %s",
+            upstream_versions[1]
         )
         logging.info(
-            "Current most recent TB Beta version:              %s", upstream_versions[3]
+            "Current most recent TB Candidate ESR (previous) version: %s build%s",
+            upstream_versions[2],
+            upstream_versions[3]
+        )
+        logging.info(
+            "Current most recent TB Candidate ESR (current) version:  %s build%s",
+            upstream_versions[4],
+            upstream_versions[5]
+        )
+        logging.info(
+            "Current most recent TB Beta version:                     %s",
+            upstream_versions[6]
+        )
+        logging.info(
+            "Current most recent non ESR Release:                     %s",
+            upstream_versions[7]
+        )
+        logging.info(
+            "Current most recent non ESR Candidate:                   %s",
+            upstream_versions[7]
         )
         sys.exit(0)
 
@@ -611,10 +755,13 @@ if __name__ == "__main__":
     do_repack(source, target, args.filter)
 
     logging.info(
-        "Thunderbird source tarball prepared as:         %s/thunderbird_%s.orig.tar.xz", tarball_path, version
+        "Thunderbird source tarball prepared as:         %s/thunderbird_%s.orig.tar.xz",
+        tarball_path, version
     )
     logging.info(
-        "Thunderbird l10n component tarball prepared as: %s/%s", tarball_path, tb_l10n_component_name
+        "Thunderbird l10n component tarball prepared as: %s/%s",
+        tarball_path,
+        tb_l10n_component_name
     )
 
     # Some checking if further adjustments are required to have packaging in
